!--------------------------
!--------------------------

#define HARDCODED_8_KBYTES_CACHES

    .text
    .global initialize_caches

! This code must be placed before first cache enable
initialize_caches:
initialize_caches_tags:
    sta %g0, [%g0] 0x02 	! Turn off cache (optional)
    flush               	! Clear VALID fields (rewrite all tags of instruction and data caches)

initialize_instrustion_cache_data_way_1:
! Note 1: STA / LDA require that the processor have been in supervisor mode (S bit of %psr must be set)
! Note 2: Instruction Cache tag/data access (ASI 0x0C-0x0D) require that Instruction Cache state must be "disabled"
! Note 3: STA / LDA to ASI 0x0C-0x0F during flush lead to data_access_exception trap

  #ifdef HARDCODED_8_KBYTES_CACHES
  !--------------------------
    set 0x00002000, %g2 	! Instruction Cache size (2 ways, 4 KByte/way, 1 line == 32 bytes, 1 word == 4 bytes)
    set 0x00001000, %g3 	! Instruction Cache single Way size
  !--------------------------
  #else
    set 0x00000008, %g3
    lda [%g3] 0x02, %g3 	! read Instruction Cache Configuration register
    srl %g3, 24, %g1    	! Instruction Cache Ways  (000 - direct mapping, 001 - 2 ways, 010 - 3 ways, 011 - 4 ways)
    and %g1, 0x7, %g1   	! mask is used
    add %g1, 1, %g1
    sll %g1, 10, %g2    	! Ways * 1 KByte
    srl %g3, 20, %g1    	! Instruction Cache WaySize (2^^WSIZE KBytes/Way)
    and %g1, 0xF, %g1   	! mask is used
    sll %g2, %g1, %g2   	! %g2 - Instruction Cache size
    set 0x00000400, %g3 	! 1 KByte
    sll %g3, %g1, %g3   	! %g3 - Instruction Cache single Way size
  !--------------------------
  #endif

    ! wait until flush will be done
    lda [%g0] 0x02, %g1
2:
    srl %g1, 14, %g1
    andcc %g1, 0x0003, %g0	! Check IP and DP bits of Cache Configuration register
    bne 2b
     lda [%g0] 0x02, %g1

    ! main write cycle
    .align 32           	! The right way is to place cycle into one prefetch block (8 instruction)
1:
    subcc %g2, 4, %g2
    !andncc %g2, %g3, %g0	! exit from cycle before addresses of Way 0 begins
    nop
    bne 1b
     sta %g0, [%g2] 0x0D	! write to ASI 0x0D (Instruction Cache data)

! Enable only instruction cache
    set 0x00000003, %g1
    sta %g1, [%g0] 0x02

initialize_data_cache_data_way_1:
! Note 1: STA / LDA require that the processor have been in supervisor mode (S bit of %psr must be set)
  #ifdef HARDCODED_8_KBYTES_CACHES
  !--------------------------
    set 0x00002000, %g2 	! Data Cache size  (2 ways, 4 KByte/way, 1 line == 16 bytes, 1 word == 4 bytes)
    set 0x00001000, %g3 	! Data Cache single Way size
  !--------------------------
  #else
  !--------------------------
    set 0x0000000C, %g3
    lda [%g3] 0x02, %g3 	! read Data Cache Configuration register
    srl %g3, 24, %g1    	! Data Cache Ways  (000 - direct mapping, 001 - 2 ways, 010 - 3 ways, 011 - 4 ways)
    and %g1, 0x7, %g1   	! mask is used
    add %g1, 1, %g1
    sll %g1, 10, %g2    	! Ways * 1 KByte
    srl %g3, 20, %g1    	! Data Cache WaySize (2^^WSIZE KBytes/Way)
    and %g1, 0xF, %g1   	! mask is used
    sll %g2, %g1, %g2   	! %g2 - Data Cache size
    set 0x00000400, %g3 	! 1 KByte
    sll %g3, %g1, %g3   	! %g3 - Data Cache single Way size
  !--------------------------
  #endif
    ! main write cycle
1:
    subcc %g2, 4, %g2
    !andncc %g2, %g3, %g0	! exit from cycle before addresses of Way 0 begins
	nop
    bne 1b
     sta %g0, [%g2] 0x0F	! write to ASI 0x0F (Data Cache data)

! Enable both caches and data cache snooping
    set 0x0080000F, %g2
    sta %g2, [%g0] 0x02

!--------------------------
!--------------------------
